REM
REM THIS FILE IS PART OF THE "GRUMPY WEB SERVER" PROJECT
REM http://www.viste-family.net/mateusz/grumpy/
REM
REM  Gopher server engine
REM  Author: Mateusz Viste <mateusz@viste-family.net>
REM  Last modified: 01 Feb 2009
REM

#INCLUDE ONCE "wordwrap.bi"  ' Provides the WordWrap() function

PRIVATE TYPE GopherLink
  Server AS STRING
  Selector AS STRING
  Type AS STRING
  Port AS STRING
  Description AS STRING
END TYPE


PRIVATE FUNCTION DetectGopherType(FilenameString AS STRING) AS STRING
  DIM GopherType AS STRING
  IF INSTR(FilenameString, ".") > 0 THEN
    SELECT CASE LCASE(MID(FilenameString, INSTRREV(FilenameString, ".")))
      CASE ".txt"
        GopherType = "0"
      CASE ".gif"
        GopherType = "g"
      CASE ".jpg", ".jpeg", ".png", ".bmp", ".pcx", ".ico", ".tif", ".tiff", ".svg", ".eps"
        GopherType = "I"
      CASE ".mp3", ".mp2", ".wav", ".mid", ".ogg", ".wma", ".flac", ".mpc", ".aiff", ".aac"
        GopherType = "s"
      CASE ".htm", ".html"
        GopherType = "h"
      CASE ".pdf"
        GopherType = "P"
      CASE ELSE
        GopherType = "9"
    END SELECT
  ELSE
    GopherType = "9"
  END IF
  RETURN GopherType
END FUNCTION


PRIVATE FUNCTION GopherSecurityCheck(GophRequest AS STRING) AS STRING
  DIM AS STRING Result = ""
  DIM AS INTEGER x
  IF LEN(GophRequest) > 512 THEN Result = "The gopher request is longer than 512 bytes. The RFC 1436 states that the selector shouldn't be longer than 256 bytes."
  IF Result = "" THEN
    IF INSTR(GophRequest, CHR(0)) > 0 THEN Result = "A null char has been found in the gopher request. There's no reason for such char to be present there."
  END IF
  IF Result = "" THEN
    FOR x = 1 TO 31
      IF INSTR(GophRequest, CHR(x)) > 0 AND x <> 9 THEN Result = "A control char (ASCII 1..31) has been found in the request. There's no reason for such char to be present there."
    NEXT x
  END IF
  IF Result = "" THEN
    IF INSTR(GophRequest, CHR(9) + CHR(9)) > 0 THEN Result = "The client's request contains two TAB characters, one after the other. It shouldn't ever happen."
  END IF
  IF Result = "" THEN
    IF RIGHT(GophRequest, 1) = CHR(9) THEN Result = "The client's request is ended by a TAB character. There's no situation where that should happen."
  END IF
  RETURN Result
END FUNCTION


PRIVATE FUNCTION ReadGopherLink(CFGfile AS STRING) AS GopherLink
 DIM AS STRING CfgReturnString, CfgTmpBuffer
 DIM AS INTEGER Counter
 DIM AS GopherLink Result
 DIM AS INTEGER CfgFileHandler, CfgColonPos

 CfgFileHandler = FREEFILE
 CfgReturnString = ""
 Counter = 0
 IF DIR(CFGfile) <> "" THEN
    OPEN CFGfile FOR INPUT AS #CfgFileHandler
    DO
       Counter += 1
       LINE INPUT #CfgFileHandler, CfgTmpBuffer
       IF MID(TRIM(CfgTmpBuffer), 1, 1) <> "#" THEN
         CfgColonPos = INSTR(CfgTmpBuffer, "=")
         Token = TRIM(MID(CfgTmpBuffer, 1, CfgColonPos - 1))
         Value = TRIM(MID(CfgTmpBuffer, CfgColonPos + 1))
         SELECT CASE UCASE(Token)
           CASE "SERVER"
             Result.Server = Value
           CASE "SELECTOR"
             Result.Selector = Value
           CASE "TYPE"
             Result.Type = Value
           CASE "PORT"
             Result.Port = Value
           CASE "DESCRIPTION"
             Result.Description = Value
         END SELECT
       END IF
    LOOP UNTIL EOF(CfgFileHandler) OR Counter = 255
    CLOSE #CfgFileHandler
 END IF
 RETURN Result
END FUNCTION



SUB GopherProcess()
  DIM AS STRING FileName, EntryDescription, GopherRoot, GopherHostname, GopherPort, GopherType, DirectoryToList, SecurityCheck, TempString, MapSelector, MapDescription, MapPort, MapServer, MapType, GopherPlusOptions
  DIM AS STRING*1 ByteBuff
  DIM AS STRING*1024 KByteBuff
  DIM AS UINTEGER FileAttrib
  DIM AS INTEGER DirIndex, x, y, SmallestEntry
  DIM AS GopherLink TempLink
  DIM DirList(1 TO 20000) AS STRING

  GopherRoot = ReadCFG("/etc/grumpy.cfg", "GopherRoot")
  GopherPort = ReadCFG("/etc/grumpy.cfg", "GopherPort")
  GopherHostname = ReadCFG("/etc/grumpy.cfg", "GopherHostname")
  IF TRIM(GopherRoot) = "" THEN GopherRoot = "/var/gopher/"
  IF TRIM(GopherPort) = "" THEN GopherPort = "70"

  DirectoryToList = LineInput(4)

  SecurityCheck = GopherSecurityCheck(DirectoryToList)
  IF SecurityCheck <> "" THEN
      IF VerboseMode > 0 THEN LogLine("The gopher security module has detected a suspect condition. The query won't be processed. Reason: " + SecurityCheck)
    ELSE
      IF INSTR(DirectoryToList, CHR(9)) > 0 THEN
        GopherPlusOptions = MID(DirectoryToList, INSTR(DirectoryToList, CHR(9)) + 1)
        DirectoryToList = MID(DirectoryToList, 1, INSTR(DirectoryToList, CHR(9)) - 1) ' Cut the selector at 1st TAB
      END IF
      IF VerboseMode > 0 THEN LogLine("GopherQuery=" + CHR(34) + DirectoryToList + CHR(34))
      DirectoryToList = RemoveDoubleSlash("/" + DirectoryToList)
      LocalFile = RemoveDoubleSlash(GopherRoot + "/" + DirectoryToList)

      IF CheckForEvasion(GopherRoot, LocalFile) = 1 THEN
          IF VerboseMode > 0 THEN LogLine("Response=""Evasion attempt. Forbidden!""")
          PRINT #5, "iForbidden!"; CHR(9); "fake"; CHR(9); "(NULL)"; CHR(9); "0"; CRLF; "."; CRLF;
        ELSE
          IF IsDirectory(LocalFile) = 1 THEN
              IF RIGHT(LocalFile, 1) <> "/" THEN LocalFile += "/"
              IF RIGHT(DirectoryToList, 1) <> "/" THEN DirectoryToList += "/"
              IF FileExists(LocalFile + "grumpy.gophermap") <> 0 THEN
                  IF VerboseMode > 0 THEN LogLine("Response=""Return gophermap. (" + LocalFile + "grumpy.gophermap)""")
                  OPEN LocalFile + "grumpy.gophermap" FOR INPUT AS #51
                  WHILE NOT EOF(51)
                    LINE INPUT #51, TempString
                    IF TempString = "i" THEN TempString = "i "
                    IF LEN(TempString) > 1 THEN
                      MapType = MID(TempString, 1, 1)
                      TempString = MID(TempString, 2)
                      x = 1 : y = 1
                      MapDescription = "" : MapSelector = "" : MapServer = "" : MapPort = ""
                      DO
                        IF MID(TempString, x, 1) = CHR(9) THEN
                            y += 1
                          ELSE
                            SELECT CASE y
                              CASE 1   ' Description
                                MapDescription += MID(TempString, x, 1)
                              CASE 2   ' Selector
                                MapSelector += MID(TempString, x, 1)
                              CASE 3   ' Server
                                MapServer += MID(TempString, x, 1)
                              CASE 4   ' Port
                                MapPort += MID(TempString, x, 1)
                            END SELECT
                        END IF
                        x += 1
                      LOOP UNTIL x > LEN(TempString) OR y > 4
                      IF TRIM(MapServer) = "" THEN MapServer = GopherHostname
                      IF TRIM(MapPort) = "" THEN MapPort = GopherPort
                      IF TRIM(MapType) = "" THEN MapType = "3" ' Error if no type specified
                      PRINT #5, MapType; MapDescription; CHR(9); MapSelector; CHR(9); MapServer; CHR(9); MapPort; CRLF;
                    END IF
                  WEND
                  CLOSE #51
                  PRINT "."; CRLF;
                ELSE
                  TempString = CheckForAuth(LocalFile)
                  IF TempString <> "" THEN
                      PRINT #5, "iThis ressource is protected by a password, and may be accessed via http only. Sorry."; CHR(9); "fake"; CHR(9); "(NULL)"; CHR(9); "0"; CRLF; "."; CRLF;
                      IF VerboseMode > 0 THEN LogLine("Response=""Password-protection. Access denied.""")
                    ELSE
                      ' LIST FILES & DIRECTORIES
                      EntryDescription = GetFileDescription(".", LocalFile)
                      IF LEN(EntryDescription) > 0 THEN
                        DO
                          TempString = WordWrap(EntryDescription, 70)
                          PRINT #5, "i"; TempString; CHR(9); "fake"; CHR(9); "(NULL)"; CHR(9); "0"; CRLF;
                        LOOP UNTIL LEN(EntryDescription) = 0
                        PRINT #5, "i"; CHR(9); "fake"; CHR(9); "NULL"; CHR(9); "0"; CRLF;
                      END IF
                      FileName = DIR(LocalFile + "*", &h37, FileAttrib)
                      DirIndex = 0
                      WHILE FileName <> "" AND DirIndex < 20000
                        IF FileName <> "." AND FileName <> ".." AND LEFT(FileName, 1) <> "." AND Filename <> "descript.ion" THEN
                          DirIndex += 1
                          DirList(DirIndex) = FileName
                          IF FileAttrib AND &h10 THEN DirList(DirIndex) = CHR(9) + DirList(DirIndex)
                        END IF
                        FileName = DIR(FileAttrib)
                      WEND
                      FOR x = 1 TO DirIndex                                    '
                        y = 0                                                  '
                        IF LEFT(DirList(x), 1) = CHR(9) THEN                   '
                          y = 1                                                '
                          DirList(x) = MID(DirList(x), 2)                      '
                        END IF                                                 '
                        TempString = ""                                        '
                        IF RIGHT(DirList(x), 11) = ".gopherlink" THEN          ' Retrieve descriptions
                            TempLink = ReadGopherLink(LocalFile + DirList(x))  ' ...
                            IF TempLink.Description = "" THEN TempLink.Description = TempLink.Server
                            IF TempLink.Description = "" AND UCASE(MID(TempLink.Selector, 1, 4)) = "URL:" THEN TempLink.Description = MID(TempLink.Selector, 5)
                            IF TempLink.Description = "" THEN TempLink.Description = LEFT(DirList(x), LEN(DirList(x)) - 11)
                            TempString = TempLink.Description
                          ELSE
                            TempString = GetFileDescription(DirList(x), LocalFile)
                            IF LEN(TempString) = 0 THEN TempString = DirList(x)
                        END IF
                        DirList(x) = TempString + CHR(9) + DirList(x)
                        IF y = 1 THEN DirList(x) = CHR(9) + DirList(x)
                      NEXT x
                    REM  ***  Here I am sorting the directory's entries ***
                      IF DirIndex > 1 THEN strQsort(DirList(), DirIndex)  ' (included in sortstr.bi)
                      ' Now, we will display all entries
                      IF LEN(FileName) > 0 THEN PRINT #5, "iWarning: Too much entries, only 20'000 are displayed!"; CHR(9); "fake"; CHR(9); "(NULL)"; CHR(9); "0"; CRLF; "i"; CHR(9); "fake"; CHR(9); "NULL"; CHR(9); "0"; CRLF;
                      IF DirIndex > 0 THEN
                          FOR x = 1 TO DirIndex
                            IF RIGHT(DirList(x), 11) = ".gopherlink" AND MID(DirList(x), 1, 1) <> CHR(9) THEN
                                TempString = MID(DirList(x), 1, INSTR(DirList(x), CHR(9)) - 1)
                                DirList(x) = MID(DirList(x), INSTR(DirList(x), CHR(9)) + 1)
                                TempLink = ReadGopherLink(LocalFile + DirList(x))
                                TempLink.Description = TempString
                                IF TempLink.Selector = "" AND TempLink.Server = "" THEN TempLink.Selector = DirectoryToList + DirList(x)
                                IF TempLink.Server = "" THEN TempLink.Server = GopherHostname
                                IF TempLink.Type = "" THEN TempLink.Type = "1"
                                IF UCASE(MID(TempLink.Selector, 1, 4)) = "URL:" THEN TempLink.Type = "h"
                                IF TempLink.Port = "" THEN TempLink.Port = "70"
                                PRINT #5, TempLink.Type; TempLink.Description; CHR(9); TempLink.Selector; CHR(9); TempLink.Server; CHR(9); TempLink.Port; CRLF;
                              ELSE
                                IF MID(DirList(x), 1, 1) = CHR(9) THEN
                                    GopherType = "1"
                                    DirList(x) = MID(DirList(x), 2)
                                  ELSE
                                    GopherType = DetectGopherType(DirList(x))
                                END IF
                                EntryDescription = MID(DirList(x), 1, INSTR(DirList(x), CHR(9)) - 1)
                                DirList(x) = MID(DirList(x), INSTR(DirList(x), CHR(9)) + 1)
                                PRINT #5, GopherType; EntryDescription; CHR(9); DirectoryToList + DirList(x); CHR(9); GopherHostname; CHR(9); GopherPort; CRLF;
                            END IF
                          NEXT x
                          PRINT "."; CRLF;
                          IF VerboseMode > 0 THEN LogLine("Response=""Return directory listing. (" + LocalFile + ")""")
                        ELSE
                          PRINT #5, "iThis directory is empty."; CHR(9); "fake"; CHR(9); "(NULL)"; CHR(9); "0"; CRLF; "."; CRLF;
                          IF VerboseMode > 0 THEN LogLine("Response=""Empty Directory""")
                      END IF
                  END IF
              END IF
            ELSE
              TempString = CheckForAuth(LocalFile)
              IF TempString <> "" THEN
                  PRINT #5, "iThis ressource is protected by a password, and may be accessed via http only. Sorry."; CHR(9); "fake"; CHR(9); "(NULL)"; CHR(9); "0"; CRLF; "."; CRLF;
                  IF VerboseMode > 0 THEN LogLine("Response=""Password-protection. Access denied.""")
                ELSE
                  IF FileExists(LocalFile) <> 0 AND RIGHT(LocalFile, 13) <> "/.grumpy.auth" THEN
                      IF VerboseMode > 0 THEN LogLine("Response=""Return file. (" + LocalFile + ")""")
                      GopherType = DetectGopherType(LocalFile)
                      SELECT CASE GopherType
                        CASE "0", "2", "6"
                          OPEN LocalFile FOR INPUT AS #20
                          WHILE NOT EOF(20)
                            LINE INPUT #20, LineBuff
                            IF LineBuff = "." THEN LineBuff = ". "
                            PRINT #5, LineBuff; CRLF;
                          WEND
                          CLOSE(20)
                          PRINT "."; CRLF;
                       CASE ELSE
                          OPEN LocalFile FOR BINARY AS #20
                          x = 1
                          WHILE LOF(20) - x > 1024        '
                            x += 1024                     ' Caching the file
                            GET #20, x - 1024, KByteBuff  ' into 1K blocks
                            PRINT #5, KByteBuff;          '
                          WEND
                          FOR x = x TO LOF(20)
                            GET #20, x, ByteBuff
                            PRINT #5, ByteBuff;
                          NEXT x
                          CLOSE(20)
                      END SELECT
                    ELSE
                      IF VerboseMode > 0 THEN LogLine("Response=""Ressource doesn't exist!""")
                      PRINT #5, "3The selected ressource doesn't exist!"; CHR(9); "fake"; CHR(9); "(NULL)"; CHR(9); "0"; CRLF;
                      PRINT #5, "iThe selected ressource cannot be located."; CHR(9); "fake"; CHR(9); "(NULL)"; CHR(9); "0"; CRLF; "."; CRLF;
                  END IF
              END IF
          END IF
      END IF
  END IF
END SUB
